from uuid import uuid4
import platform
import psutil
import ctypes
import subprocess
import sys

USER_AGENT = "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/122.0.0.0 Safari/537.36"
OMAHA_VERSION = "1.3.36.261"
CUP_PROTOCOL_VERSION = "3.1"

def _get_os() -> str:
    os_name = platform.system()
    if os_name == "Linux":
        return "linux"
    elif os_name == "Windows":
        return "win"
    elif os_name == "Darwin":
        return "mac"
    else:
        raise ValueError(f"Unsupported operating system: {os_name}")

def _get_physmemory() -> int:
    """
    Returns the physical memory of the system in GB.
    :return: The physical memory of the system in GB.
    """
    try:
        if platform.system() == "Linux":
            with open("/proc/meminfo", "r") as f:
                for line in f:
                    if "MemTotal" in line:
                        mem = int(line.split()[1]) / 1024 / 1024
                        return int(mem)
        elif platform.system() == "Windows":
            return int(psutil.virtual_memory().total / (1024 ** 3))
        elif platform.system() == "Darwin":
            result = subprocess.run(
                ["sysctl", "hw.memsize"],
                capture_output=True,
                text=True,
                check=True
            )
            mem_bytes = int(result.stdout.split(":")[1].strip())
            return int(mem_bytes / (1024 ** 3))
        else:
            raise ValueError("Unsupported operating system")
    except:
        print("Warning: Failed to retrieve physical memory. Using default value (8GB)", file=sys.stderr)
        return 8

def _get_support_flags() -> dict:
    """
    Returns a dictionary with the CPU instruction sets supported by the system.
    :return: A dictionary with the CPU instruction sets supported by the system.
    """
    FLAGS = [ "avx", "avx2", "sse", "sse2", "sse3", "sse4_1", "sse4_2", "ssse3" ]
    sse_support = { flag: False for flag in FLAGS }
    try:
        if platform.system() == "Linux":
            with open("/proc/cpuinfo", "r") as f:
                supported_flags = []
                for line in f:
                    if line.startswith("flags"):
                        supported_flags = line.split(":")[1].strip().split(" ")
                        break
                for flag in FLAGS:
                    if flag in supported_flags:
                        sse_support[flag] = True
        elif platform.system() == "Windows":
            kernel32 = ctypes.windll.kernel32
            sys_info = kernel32.GetSystemInfo()
            if sys_info.dwProcessorType == 586:
                for flag in FLAGS:
                    if flag in sys_info.dwProcessorType:
                        sse_support[flag] = True
        elif platform.system() == "Darwin":
            result = subprocess.run(
                ["sysctl", "machdep.cpu.features"],
                capture_output=True,
                text=True,
                check=True
            )
            supported_flags = result.stdout.split(":")[1].strip().split(" ")
            for flag in FLAGS:
                if flag.upper() in supported_flags:
                    sse_support[flag] = True
        else:
            raise ValueError("Unsupported operating system")
    except:
        print("Warning: Failed to retrieve CPU flags. Assuming none are active.", file=sys.stderr)

    return sse_support

def generate(component_id: str, target_version = "", send_system_info = False) -> dict:
    req_system_info = {}
    if send_system_info:
        req_system_info = {
            "arch": platform.machine(),
            "hw": {
                "physmemory": _get_physmemory(),
                **_get_support_flags()
            },
            "nacl_arch": platform.machine().replace("_", "-"),
            "os": {
                "arch": platform.machine(),
                "platform": _get_os(),
                "version": platform.version()
            },
        }

    return {
        "request": {
            "@os": _get_os(),
            "@updater": "omaha",
            "acceptformat": "crx3",
            "app": [
                {
                    "appid": component_id,
                    "enabled": True,
                    "installsource": "ondemand",
                    "lang": "en-US",
                    # "ping": {
                    #     "r": -2
                    # },
                    "updatecheck": {
                        "targetversionprefix": target_version # Select only versions from given date
                    },
                    "version": "0.0.0.0"
                }
            ],
            "dedup": "cr",
            "ismachine": False,
            "prodversion": OMAHA_VERSION,
            "protocol": CUP_PROTOCOL_VERSION,
            "requestid": f"{uuid4()}",
            "sessionid": f"{uuid4()}",
            "updaterversion": OMAHA_VERSION,
            **req_system_info
        }
    }


CHROMIUM_COMPONENT_IDS = {
  "autofill_states_data":                       "eeigpngbgcognadeebkilcpcaedhellh",
  "pki_metadata":                               "efniojlnjndmcbiieegkicadnoecjjef",
  "subresource_filter_rules":                   "gcmjkmgdlgnkkcocmoeiminaijmmjnii",
  "crowd_deny":                                 "ggkkehgbnfjpeggfpleeakpidbkibbmn",
  "certificate_error_assistant":                "giekcmmlnklenlaomppkphknjmnnpneh",
  "related_website_sets":                       "gonpemdgkjcecdgbnaabipppbmgfggbe",
  "amount_extraction_heuristic_regexes":        "hajigopbbjhghbfimgkfmpenfkclmohk",
  "crlset":                                     "hfnkpimlhhgieaddgfemjhofmfblmnib",
  "hyphenation":                                "jamhcnnkihinmdlkakkaopbjbbcngflc",
  "third_party_cookie_deprecation_metadata":    "jflhchccmppkfebkiaminageehmchikm",
  "safety_tips":                                "jflookgnkcckhobaglndicnbbgbonegd",
  "file_type_policies":                         "khaoiebndkojlmppeemjhbpbandiljpe",
  "trust_token_key_commitments":                "kiabhabjdbkjdpjbpigfodbdjmbglcoo",
  "mei_preload":                                "laoigpblnllgcgjnjnllmfolckpjlhki",
  "origin_trials":                              "llkgjffcdpffmhiakmfcdcblohccpfmo",
  "optimization_hints":                         "lmelglejhemejginpboagddgdfbepgmp",
  "cookie_readiness_list":                      "mcfjlbnicoclaecapilmleaelokfnijm",
  "screenai_library":                           "mfhmdacoffpmifoibamicehhklffanao",
  "privacy_sandbox_attestations":               "niikhdgajlphfehepabhhblakbdgeefj",
  "ondeviceheadsuggest":                        "obedbbhbpmojnkanicioggnmelmoomoc",
  "widevine_content_decryption_module":         "oimompecagnajdejgnnjijobebaeigek",
  "zxcvbn_data_dictionaries":                   "ojhpjlocmbogdgmfpkhlaaeamibhnphh",
  "open_cookie_database":                       "pmagihnlncbcefglppponlgakiphldeh"
}


import random
import base64
import hashlib
import json
import requests
from typing import Union
from .errors import DownloadFailedException, NotCrx3FileException, InvalidComponentException
from . import update_request

GOOGLE_UPDATE_URL = "https://update.googleapis.com/service/update2/json"
KEY_AMOUNT = 13
CUP_NONCE_LENGTH = 32

def _get_url(req_body_str: str) -> str:
    key_number = random.randint(1, KEY_AMOUNT)
    nonce_b64 = base64.urlsafe_b64encode(random.randbytes(CUP_NONCE_LENGTH)).decode()
    req_hash = hashlib.sha256(req_body_str.encode()).hexdigest()

    return f"{GOOGLE_UPDATE_URL}?cup2key={key_number}:{nonce_b64}&cup2hreq={req_hash}"

def _get_headers(req: dict) -> dict:
    req_updater = req["request"]['@updater']
    req_updater_version = req["request"]['prodversion']
    return {
        "x-goog-update-appid": req["request"]["app"][0]["appid"],
        "x-goog-update-interactivity": "fg",
        "x-goog-update-updater": f"{req_updater}-{req_updater_version}",
        # "user-agent": USER_AGENT,
        "content-type": "application/json",
    }

def _request_update(component_id: str, target_version = "", send_system_info = False) -> Union[tuple[str, list[str]], None]:
    req = update_request.generate(component_id, target_version, send_system_info)
    req_str = json.dumps(req)

    try:
        res = requests.post(_get_url(req_str), headers=_get_headers(req), data=req_str)
    except requests.RequestException as e:
        raise DownloadFailedException() from e
    
    try:
        res_str = res.text.lstrip(")]}'\n ")
        res_json = json.loads(res_str)
    except json.JSONDecodeError:
        raise DownloadFailedException()

    app = res_json.get("response", {}).get("app", [None])[0]
    if app is None:
        raise DownloadFailedException()
    
    app_updatecheck = app.get("updatecheck")

    if app_updatecheck is None:
        raise DownloadFailedException()

    if app_updatecheck["status"] != "ok":
        return None

    manifest = app_updatecheck.get("manifest")
    if manifest is None:
        raise DownloadFailedException()

    urls = app_updatecheck.get("urls", {}).get("url", [])
    name = manifest.get("packages", {}).get("package", [{}])[0].get("name")
    version = manifest.get("version")

    if len(urls) == 0 or version is None or name is None: 
        raise DownloadFailedException()

    return version, [ url["codebase"] + name for url in urls ]

def _get_crx3_contents(crx3: bytes) -> tuple[bytes, bytes]:
    if not crx3.startswith(b"Cr24"):
        raise NotCrx3FileException()

    version = int.from_bytes(crx3[4:8], byteorder="little")
    if version != 3:
        raise NotCrx3FileException()

    header_length = int.from_bytes(crx3[8:12], byteorder="little")

    return crx3[12:12+header_length], crx3[12+header_length:]

def _verify_header(header: bytes) -> bool:
    """
    TODO: Implement header verification
    """
    return True

def _attempt_download(url: str) -> bytes:
    res = requests.get(url)
    if res.status_code != 200:
        raise DownloadFailedException()
    
    header, content = _get_crx3_contents(res.content)
    if not _verify_header(header):
        raise InvalidComponentException()
    
    return content

def download_component(component_id: str, target_version = "", send_system_info = False) -> tuple[Union[bytes, None], Union[str, None]]:
    version, urls = _request_update(component_id, target_version, send_system_info)
    if urls is None:
        return None, None

    for url in urls:
        try:
            zip_bytes = _attempt_download(url)
            return zip_bytes, version
        except DownloadFailedException:
            continue
    raise DownloadFailedException()


import argparse
import re
import os
from .chromium_components import CHROMIUM_COMPONENT_IDS
from . import download_component
from .errors import DownloadFailedException

def validate_id(id: str):
    return bool(re.fullmatch(r"[a-z0-9]{32}", id))

def main():
    parser = argparse.ArgumentParser(description="""
    Chromium Cup Downloader CLI
    This script downloads a Chromium component using the browser's Client Update Protocol (CUP). This component is saved as a ZIP file.
    """,
    formatter_class=argparse.RawTextHelpFormatter
    )
    component_list = "\n".join(f"- {name}" for name in CHROMIUM_COMPONENT_IDS.keys())
    parser.add_argument(
        "component",
        type=str,
        help=f"The component's ID or name. The component names can be one of the following: \n{component_list}"
    )
    parser.add_argument(
        "--target_version",
        type=str,
        default="",
        help="The component's target version. It can be a prefix that the version number must match. If no value is specified, the latest version is downloaded."
    )
    parser.add_argument(
        "--output",
        type=str,
        default=None,
        help="The output file name. If no value is specified, the component's ID and version is used as the file name."
    )
    parser.add_argument(
        "--send_system_info",
        action="store_true",
        help="Send system information to the CUP server."
    )
    
    args = parser.parse_args()

    # Parse component ID: retrieve from name or use it directly
    component_name = args.component.strip().replace(" ", "_").lower()
    component_id = CHROMIUM_COMPONENT_IDS.get(component_name) or component_name
    if not validate_id(component_id):
        parser.error(f"Invalid component ID: {component_id}. The component ID must be a 32-character alphanumeric string.")

    print("Attempting to download component...")
    try:
        zipfile, downloaded_version = download_component(
            component_id,
            target_version=args.target_version,
            send_system_info=args.send_system_info
        )
        if zipfile is None or downloaded_version is None:
            parser.error(f"Component '{args.component}' (ID: {component_id}) does not have a version that matches '{args.target_version}'.")
    except DownloadFailedException:
        parser.error(f"Failed to download the component '{args.component}' (ID: {component_id}). Please check the component ID or network connection.")

    print("Component downloaded successfully. Saving to file...")
    try:
        filename = args.output or f"{component_id}-{downloaded_version}.zip"
        
        # Create output directory if it doesn't exist
        outdir = os.path.dirname(filename) or "."
        if not os.path.isdir(outdir):
            os.mkdir(outdir)
        
        # Save file
        with open(filename, "wb") as outfile:
            outfile.write(zipfile)
        
        print(f"Saved as '{filename}'.")
    except Exception as e:
        parser.error(f"Failed to save the ZIP file: {e}")

if __name__ == "__main__":
    main()